import numpy as np


def addchannel_data(data, ch_new_name, ch_new_data, section='Video'):
    """
    Add a new channel to zoo data.

    Parameters
    ----------
    data : dict
        Zoo file data.
    ch_new_name : str
        Name of the new channel.
    ch_new_data : array-like
        New data to be added to the channel (should be n x 1 or n x 3).
    section : str
        Section of zoo data ('Video' or 'Analog').

    Returns
    -------
    dict
        Updated zoo data with new channel added.

    Notes
    -----
    - If the channel already exists, it will be overwritten.
    - Adds channel name to the list in data['zoosystem'][section]['Channels'].
    """

    # Warn if overwriting
    if ch_new_name in data:
        print('Warning: channel {} already exists, overwriting...'.format(ch_new_name))

# Assign channel data
    data[ch_new_name] = {
        'line': ch_new_data,
        'event': {}
    }

    # Update channel list
    ch_list = data['zoosystem'][section].get('Channels', [])

    # If the channel list is a NumPy array, convert it to a list
    if isinstance(ch_list, np.ndarray):
        ch_list = ch_list.tolist()

    # Ensure it's a flat list of strings
    if isinstance(ch_list, list) and ch_new_name not in ch_list:
        ch_list.append(ch_new_name)
        data['zoosystem'][section]['Channels'] = ch_list

    return data


if __name__ == '__main__':
    # -------TESTING--------
    import os
    from src.biomechzoo.utils.zload import zload
    from src.biomechzoo.utils.zplot import zplot
    # get path to sample zoo file
    current_dir = os.path.dirname(os.path.abspath(__file__))
    project_root = os.path.dirname(current_dir)
    fl = os.path.join(project_root, 'data', 'other', 'HC030A05.zoo')

    # load  zoo file
    data = zload(fl)
    data = data['data']
    r = data['RKneeAngles']['line']*3
    data= addchannel_data(data, ch_new_name='blah', ch_new_data=r)
    zplot(data, 'blah')




"""
BiomechZoo: A Python toolbox for processing and analyzing human movement data.

This package provides functions for converting, processing, analyzing,
and visualizing biomechanical data (e.g., motion capture, EMG, kinetics).

Example:
    from biomechzoo import BiomechZoo
    from biomechzoo.conversion import c3d2zoo

    zoo = BiomechZoo()
    zoo.conversion.c3d2zoo('path/to/data')
"""

# Import main class or entry point
from .biomechzo import BiomechZoo

# Import commonly used submodules
from . import conversion
from . import processing
from . import plotting
from . import utils

# Define what gets exposed with "from biomechzoo import *"
__all__ = [
    "BiomechZoo",
    "conversion",
    "processing",
    "plotting",
    "utils",
]

__version__ = "0.4.4"



import numpy as np
import math
import pandas as pd
from biomechzoo.processing.add_channel_data import add_channel_data

def tilt_algorithm_data(data,ch_vert, ch_medlat, ch_antpost, plot_or_not=None):

    # extract channels from data
    avert = data[ch_vert]['line']
    amedlat = data[ch_medlat]['line']
    aantpost = data[ch_antpost]['line']

    _, avert_corr, amedlat_corr, aantpost_corr = tilt_algorithm_line(avert, amedlat, aantpost)

    data = add_channel_data(data, ch_vert + '_tilt_corr', avert_corr)
    data = add_channel_data(data, ch_medlat + '_tilt_corr', amedlat_corr)
    data = add_channel_data(data, ch_antpost + '_tilt_corr', aantpost_corr)

    return data


def tilt_algorithm_line(avert, amedlat, aantpost, plot_or_not=None):
    """
    TiltAlgorithm - to account for gravity and improper tilt alignment of a tri-axial trunk accelerometer.
    Step 1: Extract raw measured (mean) accelerations
    Step 2: Calculate tilt angles
    Step 3: Calculate horizontal dynamic accelerations vectors
    Step 4: Calculate estimated provisional vertical vector
    Step 5: Calculate vertical dynamic vector
    step 6.1:  Calculate the contribution of static components
    step 6.2 Transpose static component matrices
    step 7: Remove the static components from the templates of pre and post

    :param avert: signal predominantly in vertical direction
    :param amedlat: signal predominantly in medio-lateral direction
    :param aantpost: signal predominantly in anterior-posterior direction
    :param plot_or_not: whether to plot the results
    :return: dataframe of the tilt corrected and gravity subtracted vertical, medio-lateral and anterior-posterior
    acceleration signals
    """
    #

    a_vt = avert.mean()
    a_ml = amedlat.mean()
    a_ap = aantpost.mean()

    # if avert is negative than turn the sensor around.
    if a_vt < 0.5:
        avert *= -1
        amedlat *= -1
        a_vt = avert.mean()
        a_ml = amedlat.mean()

    # Anterior tilt
    TiltAngle_ap_rad = np.arcsin(a_ap)
    TiltAngle_ap_deg = math.degrees(TiltAngle_ap_rad)

    # mediolateral tilt
    TiltAngle_ml_rad = np.arcsin(a_ml)
    TiltAngle_ml_deg = math.degrees(TiltAngle_ml_rad)

    # Anterior posterior
    a_AP = (a_ap * np.cos(TiltAngle_ap_rad)) - (a_vt * np.sin(TiltAngle_ap_rad))
    # AMediolateral
    a_ML = (a_ml * np.cos(TiltAngle_ml_rad)) - (a_vt * np.sin(TiltAngle_ml_rad))

    # a_vt_prov = a_ap*Sin(theta_ap) + a_vt*Cos(theta_ap)
    a_vt_prov = (a_ap * np.sin(TiltAngle_ap_rad)) + (a_vt * np.cos(TiltAngle_ap_rad))

    # a_VT = a_ml*sin(theta_ml) + a_vt_prov*cos(theta_ml) - 1
    a_VT = (a_ml * np.sin(TiltAngle_ml_rad)) + (a_vt_prov * np.cos(TiltAngle_ml_rad)) - 1

    a_AP_static = a_ap - a_AP
    a_ML_static = a_ml - a_ML
    a_VT_static = a_vt - a_VT

    a_AP_static = np.transpose(a_AP_static)
    a_ML_static = np.transpose(a_ML_static)
    a_VT_static = np.transpose(a_VT_static)

    amedlat2 = amedlat - a_ML_static
    avert2 = avert - a_VT_static
    aantpost2 = aantpost - a_AP_static

    data = {'avert': avert2,
            'amedlat': amedlat2,
            'aantpost': aantpost2}
    df_corrected = pd.DataFrame(data)

    # if plot_or_not:
    #     f, ax = plt.subplots(nrows=3, ncols=1, sharex=True, dpi=300)
    #     sns.despine(offset=10)
    #     f.tight_layout()
    #     offset = 0.1
    #     f.subplots_adjust(left=0.15, top=0.95)
    #
    #     sns.lineplot(avert, ax=ax[0], label='Raw')
    #     sns.lineplot(avert2, ax=ax[0], label='tilt corrected')
    #     ax[0].set_ylabel('vert acc (g)')
    #     ax[0].set_title('Vertical acceleration corrected with {}'.format(np.round(a_VT_static, 2)))
    #
    #     sns.lineplot(amedlat, ax=ax[1], label='Raw')
    #     sns.lineplot(amedlat2, ax=ax[1], label='Tilt corrected')
    #     ax[1].set_ylabel('ml acc (g)')
    #     ax[1].set_title('Medio-lateral tilt angle corrected with {} degrees'.format(np.round(TiltAngle_ml_deg, 2)))
    #
    #     sns.lineplot(aantpost, ax=ax[2], label='Raw')
    #     sns.lineplot(aantpost2, ax=ax[2], label='Tilt corrected')
    #     ax[2].set_ylabel('ap acc (g)')
    #     ax[2].set_title('Anterior-posterior tilt angle corrected with {} degrees'.format(np.round(TiltAngle_ap_deg, 2)))

    return df_corrected, avert2, amedlat2, aantpost2


from biomechzoo.mvn import mvn
import warnings


class MvnxFileAccessor:

    @property
    def original_file_name(self):
        return self.file_data['meta_data']['original_filename']

    @property
    def actor_name(self):
        return self.file_data['meta_data']['name']

    @property
    def actor_color(self):
        return self.file_data['meta_data']['color']

    @property
    def profile(self):
        return self.file_data['meta_data']['scenario']

    @property
    def configuration(self):
        return self.file_data['meta_data']['configuration']

    @property
    def comments(self):
        return self.file_data['meta_data']['comments']

    @property
    def quality(self):
        return self.file_data['meta_data']['quality']

    @property
    def frame_count(self):
        return self._last_frame - self._first_frame

    @property
    def segment_count(self):
        return self.file_data['frames']['segment_count']

    @property
    def joint_count(self):
        return self.file_data['frames']['joint_count']

    @property
    def finger_joint_count(self):
        return self.file_data['frames']['finger_joint_count']

    @property
    def ergo_joint_count(self):
        ergo_joint_count = len(self.file_data['ergo_joints']) if self.file_data['ergo_joints'] is not None else None
        return ergo_joint_count  # 'ergoJointCount' non-existing in mvnx

    @property
    def sensor_count(self):
        return self.file_data['frames']['sensor_count']

    @property
    def frame_rate(self) -> int:
        if 'sample_rate' in self.file_data['meta_data']:
            return int(self.file_data['meta_data']['sample_rate'])
        else:
            warnings.warn('Using default sample rate of 240')
            return 240

    @property
    def recording_date(self) -> str:
        return self.file_data['meta_data']['rec_date']

    @property
    def version(self):
        return self.file_data['meta_data']['version']

    def __init__(self):
        self.file_data = {}
        self._all_frames_slice = None
        self._first_frame = 0
        self._last_frame = 0
        self._index_to_segment = {}

    def create_index_to_segment_dict(self):
        segment_index = 0
        for _, segment in self.file_data['segments']['elements'].items():
            if 'label' in segment:
                self._index_to_segment[segment_index] = segment['label']
            segment_index += 1

    def set_frame_window(self, first_frame, last_frame):
        self._first_frame = first_frame
        self._last_frame = last_frame
        self._all_frames_slice = slice(self._first_frame, self._last_frame)

    def reset_frame_window(self):
        self._first_frame = 0
        self._last_frame = len(self.file_data['frames']['segment_data'])
        self._all_frames_slice = slice(self._first_frame, self._last_frame)

    def window_profile(self):
        """
        This method determines the filter profile (aka scenario) for the currently selected window on the
        data.
        :return: The profile name (singleLevel, multiLevel, ...) or 'mixed' if different profiles
        ere applied to the window.
        """
        if 'profiles' not in self.file_data['meta_data']:
            return self.profile
        else:
            profiles = set()
            for profile_spec in self.file_data['meta_data']['profiles']:
                if profile_spec[1] > self._first_frame and profile_spec[0] <= self._last_frame:
                    profiles.add(profile_spec[2])

            if len(profiles) == 1:
                return profiles.pop()
            else:
                return 'mixed'

    def frame_to_mapped_slice(self, frame) -> (slice, bool):
        # convert the frame parameter to a slice that is mapped to the current view on the file
        single_frame = False
        if not isinstance(frame, slice):
            if frame == mvn.FRAMES_ALL:
                # all frames slice already is mapped to the current view
                self.reset_frame_window()
                frame = self._all_frames_slice  # use the prepared 'all frames' slice
            else:
                # for a single frame, map the frame to the current view on all data by adding the virtual first frame
                frame = slice(self._first_frame + frame, self._first_frame + frame + 1)  # create a single frame slice
                single_frame = True
        else:
            # shift the slice to the current virtual 'view'
            start = frame.start + self._first_frame
            if frame.stop:
                stop = frame.stop + self._first_frame
            else:  # a slice with end given as None implies end at last frame
                stop = self._last_frame
            step = frame.step
            frame = slice(start, stop, step)

        return frame, single_frame

    def segment_name_from_index(self, segment_index):
        return self._index_to_segment[segment_index]

    def point_name_from_indices(self, segment_index, point_index):
        segment_name = self.segment_name_from_index(segment_index)
        segment = self.file_data['segments']['elements'][segment_name]
        return segment['info']['point_label_from_index'][point_index]

    """ Pose methods """

    def identity_pose_is_valid(self):
        return ('identity' in self.file_data) and \
               (self.identity_pose()['segments_counts'] > 0)

    def identity_pose_segment_pos(self, segment):
        return self.identity_pose()['segments'][segment]['pos_g']

    def identity_pose_segment_ori(self, segment):
        return self.identity_pose()['segments'][segment]['q_gb']

    def t_pose_is_valid(self):
        return ('tpose' in self.file_data) and \
               (self.t_pose()['segments_counts'] > 0)

    def t_pose_segment_pos(self, segment):
        return self.t_pose()['segments'][segment]['pos_g']

    def t_pose_segment_ori(self, segment):
        return self.t_pose()['segments'][segment]['q_gb']

    def identity_pose(self):
        return self.file_data['identity']

    def t_pose(self):
        return self.file_data['tpose']

    """ Segment methods """

    def get_segment_pos(self, segment, frame=mvn.FRAMES_ALL, axis=mvn.AXIS_ALL):
        """
        Get the position information for a segment

        :param segment: The index of the segment to return the data for (Mvn.SEGMENT_.... )
        :param frame: Can be the index of the frame to return, a slice to return a range of frames
                      or Mvn.FRAMES_ALL (default) to return all frames
        :param axis: The frame number to return the data for (Mvn.AXIS_... ALL for all axes)
        :return: A single value, list, or list of lists with position values
        """

        return self.get_segment_data('pos', segment, frame, axis)

    def get_segment_ori(self, segment, frame=mvn.FRAMES_ALL, axis=mvn.AXIS_ALL):
        """
        Get the orientation information for a segment

        :param segment: The index of the segment to return the data for (Mvn.SEGMENT_.... )
        :param frame: Can be the index of the frame to return, a slice to return a range of frames
                      or Mvn.FRAMES_ALL (default) to return all frames
        :param axis: The frame number to return the data for (Mvn.AXIS_... ALL for all axes)
        :return: A single value, list, or list of lists with orientation values
        """

        # For orientation data, if all axes requested, return all. If specific axis requested, shift index so that w,
        # x,y,z becomes 0,1,2,3
        axis = (axis + 1) % 4 if axis != mvn.AXIS_ALL else axis
        return self.get_segment_data('ori', segment, frame, axis)

    def get_segment_point_pos(self, segment, point):
        segment_name = mvn.SEGMENTS[segment]
        segment_info = self.file_data['segments'][segment_name]
        points = segment_info['info']['point_label_from_index']
        point_name = points[point]
        return segment_info['points_mvn'][point_name]

    def get_point_pos(self, segment, point):
        segment_name = mvn.SEGMENTS[segment]

        if segment_name == 'LeftFoot':
            point_name = mvn.POINTS_LEFT_FOOT[point]
        elif segment_name == 'RightFoot':
            point_name = mvn.POINTS_RIGHT_FOOT[point]
        elif segment_name == 'LeftToe':
            point_name = mvn.POINTS_LEFT_TOE[point]
        elif segment_name == 'RightToe':
            point_name = mvn.POINTS_RIGHT_TOE[point]

        return self.file_data['segments'][segment_name]['points_mvn'][point_name]

    def get_segment_vel(self, segment, frame=mvn.FRAMES_ALL, axis=mvn.AXIS_ALL):
        """
        Get the local velocity information for a segment

        :param segment: The index of the segment to return the data for (Mvn.SEGMENT_.... )
        :param frame: Can be the index of the frame to return, a slice to return a range of frames
                      or Mvn.FRAMES_ALL (default) to return all frames
        :param axis: The frame number to return the data for (Mvn.AXIS_... ALL for all axes)
        :return: A single value, list, or list of lists with velocity values
        """

        return self.get_segment_data('vel', segment, frame, axis)

    def get_segment_acc(self, segment, frame=mvn.FRAMES_ALL, axis=mvn.AXIS_ALL):
        """
        Get the acceleration information for a segment

        :param segment: The index of the segment to return the data for (Mvn.SEGMENT_.... )
        :param frame: Can be the index of the frame to return, a slice to return a range of frames
                      or Mvn.FRAMES_ALL (default) to return all frames
        :param axis: The frame number to return the data for (Mvn.AXIS_... ALL for all axes)
        :return: A single value, list, or list of lists with acceleration values
        """

        return self.get_segment_data('acc', segment, frame, axis)

    def get_segment_angular_vel(self, segment, frame=mvn.FRAMES_ALL, axis=mvn.AXIS_ALL):
        """
        Get the acceleration information for a segment

        :param segment: The index of the segment to return the data for (Mvn.SEGMENT_.... )
        :param frame: Can be the index of the frame to return, a slice to return a range of frames
                      or Mvn.FRAMES_ALL (default) to return all frames
        :param axis: The frame number to return the data for (Mvn.AXIS_... ALL for all axes)
        :return: A single value, list, or list of lists with angular velocity values
        """

        return self.get_segment_data('ang_vel', segment, frame, axis)

    def get_segment_angular_acc(self, segment, frame=mvn.FRAMES_ALL, axis=mvn.AXIS_ALL):
        """
        Get the acceleration information for a segment

        :param segment: The index of the segment to return the data for (Mvn.SEGMENT_.... )
        :param frame: Can be the index of the frame to return, a slice to return a range of frames
                      or Mvn.FRAMES_ALL (default) to return all frames
        :param axis: The frame number to return the data for (Mvn.AXIS_... ALL for all axes)
        :return: A single value, list, or list of lists with acceleration values
        """

        return self.get_segment_data('ang_acc', segment, frame, axis)

    def get_segment_data(self, data_field, segment, frame=mvn.FRAMES_ALL, axis=mvn.AXIS_ALL):
        return self.get_data('segment_data', data_field, segment, frame, axis)

    """ Joint methods """

    def get_joint_angle(self, joint, frame=mvn.FRAMES_ALL, angle=mvn.ANGLE_ALL):
        joint_name = mvn.JOINTS[joint]
        data_set = 'joint_data'

        frame, is_single_frame = self.frame_to_mapped_slice(frame)

        if angle == mvn.ANGLE_ALL:
            return_values = [value[joint_name] for value in self.file_data['frames'][data_set][frame]]
        else:
            return_values = [value[joint_name][angle] for value in self.file_data['frames'][data_set][frame]]

        return return_values[0] if is_single_frame else return_values

    def get_joint_angle_xzy(self, joint, frame=mvn.FRAMES_ALL, angle=mvn.ANGLE_ALL):
        joint_name = mvn.JOINTS[joint]
        data_set = 'joint_data_xzy'

        frame, is_single_frame = self.frame_to_mapped_slice(frame)

        if angle == mvn.ANGLE_ALL:
            return_values = [value[joint_name] for value in self.file_data['frames'][data_set][frame]]
        else:
            return_values = [value[joint_name][angle] for value in self.file_data['frames'][data_set][frame]]

        return return_values[0] if is_single_frame else return_values

    def get_ergo_joint_angle(self, joint, frame=mvn.FRAMES_ALL, angle=mvn.ANGLE_ALL):
        joint_name = mvn.ERGO_JOINTS[joint]
        data_set = 'ergo_joint_data'

        frame, is_single_frame = self.frame_to_mapped_slice(frame)

        if angle == mvn.ANGLE_ALL:
            return_values = [value[joint_name] for value in self.file_data['frames'][data_set][frame]]
        else:
            return_values = [value[joint_name][angle] for value in self.file_data['frames'][data_set][frame]]

        return return_values[0] if is_single_frame else return_values

    """ Center of Mass methods """

    def get_center_of_mass_pos(self, frame=mvn.FRAMES_ALL, axis=mvn.AXIS_ALL):
        """
        Get the position information for center of mass

        :param frame: Can be the index of the frame to return, a slice to return a range of frames
                      or Mvn.FRAMES_ALL (default) to return all frames
        :param axis: The axis to return the data for (Mvn.AXIS_... ALL for all axes)
        :return: A single value, list, or list of lists with position values
        """

        return self.get_segment_data('pos', mvn.SEGMENT_CENTER_OF_MASS, frame, axis)

    def get_center_of_mass_vel(self, frame=mvn.FRAMES_ALL, axis=mvn.AXIS_ALL):
        """
        Get the velocity information for center of mass

        :param frame: Can be the index of the frame to return, a slice to return a range of frames
                      or Mvn.FRAMES_ALL (default) to return all frames
        :param axis: The axis to return the data for (Mvn.AXIS_... ALL for all axes)
        :return: A single value, list, or list of lists with position values
        """

        return self.get_segment_data('vel', mvn.SEGMENT_CENTER_OF_MASS, frame, axis)

    def get_center_of_mass_acc(self, frame=mvn.FRAMES_ALL, axis=mvn.AXIS_ALL):
        """
        Get the acceleration information for center of mass

        :param frame: Can be the index of the frame to return, a slice to return a range of frames
                      or Mvn.FRAMES_ALL (default) to return all frames
        :param axis: The axis to return the data for (Mvn.AXIS_... ALL for all axes)
        :return: A single value, list, or list of lists with position values
        """

        return self.get_segment_data('acc', mvn.SEGMENT_CENTER_OF_MASS, frame, axis)

    """ Sensor methods """

    def get_sensor_ori(self, segment, frame=mvn.FRAMES_ALL, axis=mvn.AXIS_ALL):
        """
        Get the orientation information for a sensor

        :param segment: The index of the segment to return the sensor data for (Mvn.SEGMENT_.... )
        :param frame: Can be the index of the frame to return, a slice to return a range of frames
                      or Mvn.FRAMES_ALL (default) to return all frames
        :param axis: The frame number to return the data for (Mvn.AXIS_... ALL for all axes)
        :return: A single value, list, or list of lists with orientation values
        """

        # if all axes requested, return all. If specific axis requested, shift index so that w,x,y,z becomes 0,1,2,3
        axis = (axis + 1) % 4 if axis != mvn.AXIS_ALL else axis
        return self.get_sensor_data('ori', segment, frame, axis)

    def get_sensor_free_acc(self, segment, frame=mvn.FRAMES_ALL, axis=mvn.AXIS_ALL):
        """
        Get the free acceleration information for a sensor

        :param segment: The index of the segment to return the sensor data for (Mvn.SEGMENT_.... )
        :param frame: Can be the index of the frame to return, a slice to return a range of frames
                      or Mvn.FRAMES_ALL (default) to return all frames
        :param axis: The frame number to return the data for (Mvn.AXIS_... ALL for all axes)
        :return: A single value, list, or list of lists with acceleration values
        """

        return self.get_sensor_data('acc', segment, frame, axis)

    def get_sensor_data(self, data_field, sensor_segment, frame=mvn.FRAMES_ALL, axis=mvn.AXIS_ALL):
        return self.get_data('sensor_data', data_field, sensor_segment, frame, axis)

    """ Contact methods """

    def get_foot_contacts(self, frame):
        """
        Get the contacts for a frame
        :param frame: The frame, or a range of frames to retrieve the contacts for.
        :return: The contacts
        """
        frame, is_single_frame = self.frame_to_mapped_slice(frame)
        return_values = self.file_data['frames']['contacts_data'][frame] #edited Phil Dixon
        return return_values[0] if is_single_frame else return_values


    def has_foot_contact(self, frame, foot_contact_flags=0):
        """
        Find out if the frame has a contact, optionally for a specific segment/point combo

        :param frame: The frame (or range) to retrieve the contacts for.
        :param foot_contact_flags: The specific contact to check for:
                                    mvn.FOOT_CONTACT_LEFT_HEEL,
                                    mvn.FOOT_CONTACT_LEFT_TOE,
                                    mvn.FOOT_CONTACT_RIGHT_HEEL or
                                    mvn.FOOT_CONTACT_RIGHT_TOE

                                 It is possible to combine contacts by summing the values, eg:
                                    mvn.FOOT_CONTACT_LEFT_HEEL + mvn.FOOT_CONTACT_LEFT_TOE
                                 This will return a contact if either of the flags has a contact

                                 Passing 0 (or nothing) will return True if there is any contact

        :return: Per frame True if a contact was found, False otherwise.
        """
        frame_contacts = self.get_foot_contacts(frame)  # frame will be shifted in the called method

        if isinstance(frame_contacts, int):
            if foot_contact_flags == 0:
                return frame_contacts > 0
            else:
                return True if (frame_contacts & foot_contact_flags) > 0 else False
        else:
            has_contacts = []
            for contacts in frame_contacts:
                if foot_contact_flags == 0:
                    has_contacts.append(frame_contacts > 0)
                else:
                    has_contacts.append(True if (contacts & foot_contact_flags) > 0 else False)
            return has_contacts

    """ Generic methods """

    def get_data(self, data_set, data_field, segment, frame=mvn.FRAMES_ALL, axis=mvn.AXIS_ALL):
        if segment == -1:
            segment_name = 'com'
        else:
            segment_name = mvn.SEGMENTS[segment]

        frame, is_single_frame = self.frame_to_mapped_slice(frame)
        if data_set == 'joint_data':
            return_values = [value[data_field] for value in self.file_data['frames'][data_set][frame]]
        elif axis == mvn.AXIS_ALL:
            return_values = [value[segment_name][data_field] for value in self.file_data['frames'][data_set][frame]]
        else:
            return_values = [value[segment_name][data_field][axis] for value in
                             self.file_data['frames'][data_set][frame]]

        return return_values[0] if is_single_frame else return_values



def removechannel_data(data, channels, mode='remove'):
    """
    File-level processing: Remove or keep specified channels in a single zoo dictionary.

    Parameters:
    - data (dict): Zoo data loaded from a file
    - channels (list of str): List of channels to remove or keep
    - mode (str): 'remove' or 'keep'

    Returns:
    - dict: Modified zoo dictionary with updated channels
    """
    if mode not in ['remove', 'keep']:
        raise ValueError("mode must be 'remove' or 'keep'.")

    all_channels = [ch for ch in data if ch != 'zoosystem']

    # Check for missing channels
    missing = [ch for ch in channels if ch not in all_channels]
    if missing:
        print('Warning: the following channels were not found {}'.format(missing))

    if mode == 'remove':
        keep_channels = [ch for ch in all_channels if ch not in channels]
    elif mode == 'keep':
        keep_channels = [ch for ch in all_channels if ch in channels]
    else:
        raise ValueError("Mode must be 'remove' or 'keep'.")

    # --- Compute channels to remove ---
    remove_channels = [ch for ch in all_channels if ch not in keep_channels]

    if remove_channels:
        print('Removing channels: {}'.format(remove_channels))
    else:
        print('No channels to remove')

    # Remove from main data dict ---
    for ch in remove_channels:
        data.pop(ch, None)
        if ch in data['zoosystem']['Video']['Channels']:
            data['zoosystem']['Video']['Channels'] = [c for c in data['zoosystem']['Video']['Channels'] if c != ch]
        if ch in data['zoosystem']['Analog']['Channels']:
            data['zoosystem']['Analog']['Channels'] = [c for c in data['zoosystem']['Analog']['Channels'] if c != ch]

    return data



import numpy as np

def peak_sign(r):
    """
    Determine whether the largest absolute peak in the signal is positive or negative.

    Parameters
    ----------
    r : array-like
        Signal vector.

    Returns
    -------
    sign : int
        1 if the maximum peak is positive, -1 if negative.
    """
    r = np.asarray(r)
    max_val = np.max(r)
    min_val = np.min(r)

    if abs(max_val) > abs(min_val):
        return 1
    else:
        return -1



import numpy as np
import copy
import warnings
from biomechzoo.utils.peak_sign import peak_sign
from biomechzoo.biomech_ops.movement_onset import movement_onset

def addevent_data(data, channels, ename, etype, constant=None):

    data_new = copy.deepcopy(data)

    if isinstance(channels, str):
        channels = [channels]

    if len(channels) == 1 and channels[0].lower() == 'all':
        channels = [key for key in data if key != 'zoosystem']

    for channel in channels:
        if ename == '':
            data[channel]['event'] = {}
            continue

        if channel not in data:
            raise KeyError('Channel {} does not exist'.format(channel))

        yd = data_new[channel]['line']  # 1D array
        etype = etype.lower()
        if etype == 'absmax':
            exd = int(np.argmax(np.abs(yd)))
            eyd = float(yd[exd])
        elif etype == 'first':
            exd = 0
            eyd = float(yd[exd])
        elif etype == 'last':
            exd = len(yd) - 1
            eyd = float(yd[exd])
        elif etype == 'max':
            exd = int(np.argmax(yd))
            eyd = float(yd[exd])
        elif etype == 'min':
            exd = int(np.argmin(yd))
            eyd = float(yd[exd])
        elif etype == 'rom':
            eyd = float(np.max(yd) - np.min(yd))
            exd = 0  # dummy index (like MATLAB version)
        elif etype == 'max_stance':
            # special event for gait and running
            exd = max_stance(yd)
            eyd = float(yd[exd])
            eyd = float(yd[exd])
        elif etype == 'movement_onset':
            exd = movement_onset(yd, constant)
            eyd = yd[exd]
        elif etype == 'movement_offset':
            exd = movement_onset(yd,constant)
            eyd = yd[exd]
        elif etype in ['fs_fp', 'fo_fp']:
            # --- Handle constant ---
            if constant is None:
                print('Warning: Force plate threshold not set, defaulting to 0.')
                constant = 0.0

            # --- Check sampling rates ---
            AVR = data['zoosystem']['AVR']
            if AVR != 1:
                warnings.warn('Video and Analog channels must be at the same sampling rate or events will be incorrect.')

            # --- Handle units ---
            units = data['zoosystem']['Units']['Forces']
            if units == 'N/kg':
                m = data['zoosystem']['Anthro']['Bodymass']
            else:
                m = 1.0

            # --- Extract force signal ---
            if '_' not in channel:
                yd = data_new[channel]['line'][:, 2]  # looking for GRF Z
            else:
                yd = data_new[channel]['line']

            # --- Determine peak sign ---
            peak = peak_sign(yd)  # user-defined function

            # --- Find threshold crossing ---
            threshold_signal = peak * yd * m
            if 'fs' in etype:
                exd_array = np.where(threshold_signal > constant)[0]
                exd = exd_array[0] - 1  # MATLAB indexing correction
                eyd = yd[exd]
            else:  # 'FO' type
                exd_array = np.where(threshold_signal > constant)[0]
                exd = exd_array[-1] + 1
                eyd = yd[exd]

        else:
            raise ValueError(f'Unknown event type: {etype}')

        # Add event to the channel's event dict
        data_new[channel]['event'][ename] = [exd, eyd, 0]

    return data_new

def max_stance(yd):
    """ extracts max from first 40% of the gait cycle"""
    raise NotImplementedError
    return exd


import numpy as np
import scipy.signal as sgl


def filter_line(signal_raw, filt=None, fs=None):
    """Filter an array using a Butterworth filter."""
    #todo: verify that filter is working correctly
    #todo add more filters
    #todo: consider using kineticstoolkit

    if filt is None:
        filt = {'ftype': 'butter',
                'order': 4,
                'cutoff': 10,
                'btype': 'lowpass',
                'filtfilt': True}
        if fs is None:
            raise ValueError('fs is required if no filt is specified')

    else:
        if 'fs' not in filt:
            raise ValueError('fs is a required key of filt')

    # Normalize filter type strings
    if filt['ftype'] == 'butterworth':
        filt['ftype'] = 'butter'
    if filt['btype'] is 'low':
        filt['btype'] = 'lowpass'
    if filt['btype'] is 'high':
        filt['btype'] = 'highpass'

    # Extract parameters
    ftype = filt['ftype']
    order = filt['order']
    cutoff = filt['cutoff']
    btype = filt['btype']
    filtfilt = filt['filtfilt']
    fs = filt['fs']

    # prepare normalized cutoff(s)
    nyq = 0.5 * fs
    norm_cutoff = np.atleast_1d(np.array(cutoff) / nyq)

    if ftype is 'butter':
        signal_filtered = kt_butter(ts=signal_raw, fc=norm_cutoff, fs=fs, order=order, btype=btype, filtfilt=filtfilt)
    else:
        raise NotImplementedError(f"Filter type '{ftype}' not implemented.")

    return signal_filtered


def kt_butter(ts, fc, fs, order=2, btype='lowpass', filtfilt=True):
    """
    Apply a Butterworth filter to data.

    Parameters
    ----------
    ts, ndarray, 1d.
    fc, Cut-off frequency in Hz. This is a float for single-frequency filters
        (lowpass, highpass), or a tuple of two floats (e.g., (10., 13.)
        for two-frequency filters (bandpass, bandstop)).
    order, Optional. Order of the filter. Default is 2.
    btype, Optional. Can be either "lowpass", "highpass", "bandpass" or
        "bandstop". Default is "lowpass".
    filtfilt, Optional. If True, the filter is applied two times in reverse direction
        to eliminate time lag. If False, the filter is applied only in forward
        direction. Default is True.

    Returns
    -------
    ts_f,  A copy of the input data which each data being filtered.
    
    Notes: 
    - This code was adapted from kineticstoolkit Thanks @felxi
    """

    sos = sgl.butter(order, fc, btype, analog=False, output="sos", fs=fs)

    # Filter
    if filtfilt:
        ts_f = sgl.sosfiltfilt(sos, ts, axis=0)
    else:
        ts_f = sgl.sosfilt(sos,ts, axis=0)

    return ts_f


from scipy.io import loadmat
import os
import numpy as np


def zload(filepath):
    if not filepath.endswith('.zoo'):
        raise ValueError(f"{filepath} is not a .zoo file")

    if not os.path.exists(filepath):
        raise FileNotFoundError(f"File not found: {filepath}")

    mat_data = loadmat(filepath, struct_as_record=False, squeeze_me=True)

    # Remove default MATLAB metadata fields
    mat_data = {k: v for k, v in mat_data.items() if not k.startswith('__')}

    # Convert MATLAB structs to Python dicts (recursively)
    def mat_struct_to_dict(obj):
        if isinstance(obj, dict):
            return {k: mat_struct_to_dict(v) for k, v in obj.items()}
        elif hasattr(obj, '_fieldnames'):
            return {field: mat_struct_to_dict(getattr(obj, field)) for field in obj._fieldnames}
        elif isinstance(obj, list):
            return [mat_struct_to_dict(item) for item in obj]
        else:
            return obj
    data = {k: mat_struct_to_dict(v) for k, v in mat_data.items()}

    if 'data' in data:
        data = data['data']

    # Convert Video and Analog channel arrays to Python lists
    for sys in ['Video', 'Analog']:
        if 'zoosystem' in data and sys in data['zoosystem']:
            channels = data['zoosystem'][sys].get('Channels', [])
            # Convert to list and strip spaces
            if isinstance(channels, np.ndarray):
                channels = channels.tolist()
            channels = [str(ch).strip() for ch in channels]
            data['zoosystem'][sys]['Channels'] = channels

    return data


if __name__ == '__main__':
    """ testing: load a single zoo file from the other subfolder in data"""
    # -------TESTING--------
    current_dir = os.path.dirname(os.path.abspath(__file__))
    project_root = os.path.dirname(current_dir)
    fl = os.path.join(project_root, 'data', 'other', 'HC030A05.zoo')
    data = zload(fl)

    channels = [k for k in data.keys()]
    print('{} channels found'.format(len(channels)))
    for ch in channels:
        print({ch})



import copy
from biomechzoo.utils.findfield import findfield


def split_trial(data, start_event, end_event):
    # todo check index problem compared to matlab start at 0 or 1
    data_new = copy.deepcopy(data)

    start_event_indx, _ = findfield(data_new, start_event)
    end_event_indx, _ = findfield(data_new, end_event)

    for key, value in data_new.items():
        if key == 'zoosystem':
            continue

        # Slice the line data
        trial_length = len(data_new[key]['line'])
        if trial_length > end_event_indx[0]:
            data_new[key]['line'] = value['line'][start_event_indx[0]:end_event_indx[0]]
        else:
            print('skipping split trial since event is outside range of data')
            return None

        # Update events if present
        # if 'event' in value:
        #     new_events = {}
        #     for evt_name, evt_val in value['event'].items():
        #         event_frame = evt_val[0]
        #         # Check if event falls within the new window
        #         if start_event_indx <= event_frame < end_event_indx:
        #             # Adjust index relative to new start
        #             new_events[evt_name] = [event_frame - start_event_indx, 0, 0]
        #     data_new[key]['event'] = new_events

    return data_new



import copy
import warnings
from biomechzoo.utils.findfield import findfield

def removeevent_data(data, events, mode='remove'):
    """
    Remove or keep specified events in all channels of a zoo dictionary.

    Parameters
    ----------
    data : dict
        Zoo data loaded from a file
    events : list of str
        Events to remove or keep
    mode : str
        'remove' or 'keep'

    Returns
    -------
    dict
        Modified zoo dictionary with events removed or kept
    """
    if mode not in ['remove', 'keep']:
        raise ValueError("mode must be 'remove' or 'keep'.")

    if isinstance(events, str):
        events = [events]

    # check if any events are not present
    valid_events = []
    for evt in events:
        e, _ = findfield(data, evt)
        if e is None:
            warnings.warn('Could not find event {} in zoo file, skipping'.format(evt))
        else:
            valid_events.append(evt)
    events = valid_events

    data_new = copy.deepcopy(data)
    channels = sorted([ch for ch in data_new if ch != 'zoosystem'])
    for ch in channels:
        event_dict = data_new[ch].get('event', {})
        events_to_remove = []

        for evt in list(event_dict.keys()):
            if mode == 'remove' and evt in events:
                events_to_remove.append(evt)
            elif mode == 'keep' and evt not in events:
                events_to_remove.append(evt)

        for evt in events_to_remove:
            event_dict.pop(evt, None)
            # print('Removed event "{}" from channel "{}"'.format(evt, ch))

        data_new[ch]['event'] = event_dict

    return data_new



"""
Script to load an mvnx

"""

import xml.etree.ElementTree as ET
import collections
import numpy as np
from biomechzoo.mvn.mvnx_file_accessor import MvnxFileAccessor
from biomechzoo.mvn import mvn

# Xml namespace for mvnx files
ns = {'mvn': 'http://www.xsens.com/mvn/mvnx'}

# Map for easier conversion of foot contacts
FOOT_CONTACT_MAP = {'LeftFoot_Heel': {'type': 1, 'segment_index': mvn.SEGMENT_LEFT_FOOT, 'point_index': 2},
                    'LeftFoot_Toe': {'type': 1, 'segment_index': mvn.SEGMENT_LEFT_TOE, 'point_index': 1},
                    'RightFoot_Heel': {'type': 1, 'segment_index': mvn.SEGMENT_RIGHT_FOOT, 'point_index': 2},
                    'RightFoot_Toe': {'type': 1, 'segment_index': mvn.SEGMENT_RIGHT_TOE, 'point_index': 1}}


def load_mvnx(file_name):
    """
    This function opens and reads the file as an mvnx formatted XML file

    :param file_name: Name of the file to open, must have full path and .mvnx extension
    :returns: A dictionary with the data from the mvnx file
    """

    mvnx_file = MvnxFileAccessor()
    init_file_data(mvnx_file)
    tree = ET.parse(file_name)
    root = tree.getroot()

    # Get the version
    mvnx_file.file_data['meta_data']['version'] = root.get('version')

    # Find the comment element
    comment_element = root.find('mvn:comment', ns)
    if comment_element is not None:
        mvnx_file.file_data['meta_data']['comments'] = comment_element.text

    # Find the subject element
    subject_element = root.find('mvn:subject', ns)
    mvnx_file.file_data['meta_data']['name'] = subject_element.get('label')
    mvnx_file.file_data['meta_data']['color'] = subject_element.get('torsoColor')
    mvnx_file.file_data['meta_data']['sample_rate'] = subject_element.get('frameRate')
    mvnx_file.file_data['meta_data']['rec_date'] = subject_element.get('recDate')
    mvnx_file.file_data['meta_data']['original_filename'] = subject_element.get('originalFilename')
    mvnx_file.file_data['meta_data']['configuration'] = subject_element.get('configuration')
    mvnx_file.file_data['meta_data']['scenario'] = subject_element.get('userScenario')
    mvnx_file.file_data['meta_data']['quality'] = subject_element.get('processingQuality')

    # Parse the segments and their points
    segments_element = subject_element.find('mvn:segments', ns)
    segment_elements = segments_element.findall('mvn:segment', ns)
    mvnx_file.file_data['segments'] = parse_segments(segment_elements)
    mvnx_file.create_index_to_segment_dict()  # for later convenience on retrieving segment names

    # Parse sensor information
    sensors_element = subject_element.find('mvn:sensors', ns)
    if sensors_element is not None:
        mvnx_file.file_data['sensors'] = parse_sensor(sensors_element, mvnx_file.file_data['segments']['names'])

    # Parse joint information
    joints_element = subject_element.find('mvn:joints', ns)
    if joints_element is not None:
        mvnx_file.file_data['joints'] = parse_joints(joints_element, mvnx_file.file_data['segments'])

    # Parse ergo joint information
    ergo_joints_element = subject_element.find('mvn:ergonomicJointAngles', ns)
    if ergo_joints_element is not None:
        mvnx_file.file_data['ergo_joints'] = parse_ergo_joints(ergo_joints_element)

    # Parse foot contact
    foot_contact_definitions_element = subject_element.find('mvn:footContactDefinition', ns)
    if foot_contact_definitions_element is not None:
        foot_contact_definition_elements = foot_contact_definitions_element.findall('mvn:contactDefinition', ns)
        for foot_contact_definition_element in foot_contact_definition_elements:
            contact_label = foot_contact_definition_element.get('label')
            contact_index = int(foot_contact_definition_element.get('index'))
            mvnx_file.file_data['foot_contact_def'][contact_index] = FOOT_CONTACT_MAP[contact_label]

    # Parse the finger segments and their points
    for side in mvnx_file.file_data['finger_segments']['elements']:
        finger_segment_elements = subject_element.find('mvn:fingerTrackingSegments' + side.capitalize(), ns)
        if finger_segment_elements is not None:
            finger_segments = parse_segments(finger_segment_elements)
            mvnx_file.file_data['finger_segments']['elements'][side] = finger_segments['elements']
            mvnx_file.file_data['finger_segments']['names'][side] = finger_segments['names']

    # Parse finger joint information
    for side in mvnx_file.file_data['finger_joints']['elements']:
        finger_joints_element = subject_element.find('mvn:fingerTrackingJoints' + side.capitalize(), ns)
        if finger_joints_element is not None:
            finger_segments = {'names': mvnx_file.file_data['finger_segments']['names'][side],
                               'elements': mvnx_file.file_data['finger_segments']['elements'][side]}
            finger_joints = parse_joints(finger_joints_element, finger_segments)
            mvnx_file.file_data['finger_joints']['elements'][side] = finger_joints['elements']
            mvnx_file.file_data['finger_joints']['names'][side] = finger_joints['names']

    # At last, parse the actual frames
    frames_element = subject_element.find('mvn:frames', ns)
    mvnx_file.file_data['frames'], mvnx_file.file_data['tpose'],\
        mvnx_file.file_data['tpose_isb'], mvnx_file.file_data['identity'] = parse_frames(frames_element, mvnx_file)

    mvnx_file.reset_frame_window()  # Reset window so frame count displays total frame count
    return mvnx_file


def parse_sensor(sensors_element, segment_names):
    """
    Parse the sensor element

    :param sensors_element: The joint element to parse
    :param segment_names: a list with the segment names
    :return: a dictionary with sensor data indexed by sensor name and a list with the sensor names
    """
    sensor_elements = sensors_element.findall('mvn:sensor', ns)
    sensor_number = 0
    sensors = {}
    sensor_names = []
    for sensor_element in sensor_elements:
        sensor_name = sensor_element.get('label')
        sensor_names.append(sensor_name)
        sensor = {'type': 'Sensor',
                  'label': sensor_name,
                  'info': {
                      'sensor_number': sensor_number,
                      'sensor_location': segment_names.index(sensor_name) + 1}}

        sensors[sensor_name] = sensor
        sensor_number += 1
    return {'names': sensor_names, 'elements': sensors}


def parse_joints(joints_element, segments):
    """
    Parse the joint element

    :param joints_element: The joint element to parse
    :param segments: The dictionary with segment data
    :return: a dictionary with joint data indexed by joint name and a list with the joint names
    """
    joint_elements = joints_element.findall('mvn:joint', ns)

    joints = []
    joint_names = []

    for joint_element in joint_elements:
        joint = {'label': joint_element.get('label')}
        joint_names.append(joint['label'])

        segment1_index, point1_index = get_connector_indices(joint_element, 'mvn:connector1', segments)
        segment2_index, point2_index = get_connector_indices(joint_element, 'mvn:connector2', segments)
        joint['seg_points'] = np.array([[segment1_index, point1_index], [segment2_index, point2_index]])
        joints.append(joint)

    return {'names': joint_names, 'elements': joints}


def parse_ergo_joints(ergo_joints_element):
    """
    Parse the ergo joint element

    :param ergo_joints_element: The joint element to parse
    :return: a dictionary with ergo joint data indexed by joint name and a list with the ergo joint names
    """
    ergo_joint_elements = ergo_joints_element.findall('mvn:ergonomicJointAngle', ns)

    ergo_joints = []
    ergo_joint_names = []
    for ergo_joint_index in range(len(ergo_joint_elements)):
        ergo_joint_element = ergo_joint_elements[ergo_joint_index]
        ergo_joint = {'label': ergo_joint_element.get('label'),
                      'index': ergo_joint_index,
                      'parent_segment': ergo_joint_element.get('parentSegment'),
                      'child_segment': ergo_joint_element.get('childSegment')}
        ergo_joint_names.append(ergo_joint['label'])
        ergo_joints.append(ergo_joint)

    return {'names': ergo_joint_names, 'elements': ergo_joints}


def parse_segments(segment_elements):
    """
    Parse the segment element

    :param segment_elements: The segment element to parse
    :return: a dictionary a list with the segment names and segment data indexed by segment name
    """
    segments = collections.OrderedDict()
    segment_names = []

    for segment_index in range(len(segment_elements)):
        segment_element = segment_elements[segment_index]
        segment_name = segment_element.get('label')
        segment_names.append(segment_name)
        segment = {'points_mvn': collections.OrderedDict(),
                   'type': 'Segment',
                   'info': {},
                   'label': segment_name}

        info = {'segment_number': segment_index,
                'point_label_from_index': {},
                'point_origin': '',
                'adjacent_joints': []}

        points_element = segment_element.find('mvn:points', ns)
        point_elements = points_element.findall('mvn:point', ns)

        point_labels_from_index = {}
        for point_index in range(len(point_elements)):
            point_element = point_elements[point_index]

            point_name = point_element.get('label')
            point_labels_from_index[point_index] = point_name

            pos_b = point_element.find('mvn:pos_b', ns)
            segment['points_mvn'][point_name] = np.array([float(pos) for pos in pos_b.text.split(' ')])

            # if the point offset is really small, consider it the origin (could just pick 1st or all 0's as well)
            if np.sqrt(np.sum([sq * sq for sq in segment['points_mvn'][point_name]])) < 0.00001:
                info['point_origin'] = point_name

            # wild guess...
            if point_name[0] == 'j':
                info['adjacent_joints'].append(point_name)

        info['point_label_from_index'] = point_labels_from_index
        segment['info'] = info
        segments[segment_name] = segment

    return {'names': segment_names, 'elements': segments}


def get_connector_indices(joint_element, connector, segments):
    connector_element = joint_element.find(connector, ns)
    tokens = connector_element.text.split('/')
    segment_index = segments['names'].index(tokens[0])
    point_index = -1
    for key, value in segments['elements'][tokens[0]]['info']['point_label_from_index'].items():
        if value == tokens[1]:
            point_index = key
            break

    return segment_index, point_index


def parse_frames(frames_element, mvnx_file):
    """
    Parse the frames element

    :param frames_element: The frames element to parse
    :param mvnx_file: a dictionary containing, among others, a list of names
    :return: a dictionary with frames data
    """
    frames = {'time': [],
              'segment_data': [],
              'sensor_data': [],
              'joint_data': [],
              'joint_data_xzy': [],
              'ergo_joint_data': [],
              'ergo_joint_data_xzy': [],
              'contacts_data': [],
              'finger_segment_data': [],
              'finger_joint_data': {'left': [], 'right': []},
              'finger_joint_data_xzy': {'left': [], 'right': []}}
    tpose = {}
    tpose_isb = {}
    identity = {}

    get_count = lambda element: int(element) if element is not None else None
    frames['segment_count'] = get_count(frames_element.get('segmentCount'))
    frames['sensor_count'] = get_count(frames_element.get('sensorCount'))
    frames['joint_count'] = get_count(frames_element.get('jointCount'))
    frames['finger_joint_count'] = get_count(frames_element.get('fingerJointCount'))
    frame_elements = frames_element.findall('mvn:frame', ns)
    for frame_element in frame_elements:
        if frame_element.get('type') == 'normal':
            frames['time'].append(frame_element.get('time'))
            frames['joint_data'].append(
                get_joint_data_from_frame(frame_element, 'jointAngle', mvnx_file.file_data['joints']['names']))
            frames['joint_data_xzy'].append(
                get_joint_data_from_frame(frame_element, 'jointAngleXZY', mvnx_file.file_data['joints']['names']))
            frames['ergo_joint_data'].append(
                get_joint_data_from_frame(frame_element, 'jointAngleErgo', mvnx_file.file_data['ergo_joints']['names']))
            frames['ergo_joint_data_xzy'].append(
                get_joint_data_from_frame(frame_element, 'jointAngleErgoXZY', mvnx_file.file_data['ergo_joints']['names']))
            frames['segment_data'].append(
                get_segment_data_from_frame(frame_element, mvnx_file.file_data['segments']['names']))
            frames['sensor_data'].append(
                get_sensor_data_from_frame(frame_element, mvnx_file.file_data['sensors']['names']))
            frames['contacts_data'].append(
                get_contact_data_from_frame(frame_element, mvnx_file.file_data['foot_contact_def']))
            frames['finger_segment_data'].append(
                get_finger_data_from_frame(frame_element, mvnx_file.file_data['finger_segments']['names']))
            for side in frames['finger_joint_data']:
                element_name = 'jointAngleFingers' + side.capitalize()
                frames['finger_joint_data'][side].append(get_joint_data_from_frame(
                    frame_element, element_name, mvnx_file.file_data['finger_joints']['names'][side]))
                element_name = 'jointAngleFingers' + side.capitalize() + 'XZY'
                frames['finger_joint_data_xzy'][side].append(get_joint_data_from_frame(
                    frame_element, element_name, mvnx_file.file_data['finger_joints']['names'][side]))
        elif frame_element.get('type') == 'tpose':
            tpose = get_t_pose_data_from_frame(frame_element, mvnx_file.file_data['segments']['names'])
        elif frame_element.get('type') == 'tpose-isb':
            tpose_isb = get_t_pose_data_from_frame(frame_element, mvnx_file.file_data['segments']['names'])
        elif frame_element.get('type') == 'identity':
            identity = get_t_pose_data_from_frame(frame_element, mvnx_file.file_data['segments']['names'])

    return frames, tpose, tpose_isb, identity


def get_joint_data_from_frame(frame_element, joint_element_name, joint_names):
    """
    Extract joint data from a frame

    :param frame_element: The frame element to process
    :param joint_element_name: The name of the frame element to process
    :param joint_names: a list with the joint names
    :return: a dictionary with joint data indexed by joint name
    """

    joint_data = collections.OrderedDict()

    angles = frame_element_as_floats(frame_element, joint_element_name)

    for index in range(len(joint_names)):
        joint_data[joint_names[index]] = get_3d_vector(angles, index)

    return joint_data


def get_t_pose_data_from_frame(frame_element, segment_names):
    """
    Extract segment data from a frame

    :param frame_element: The frame element to process
    :param segment_names: a list with the segment names
    :return: a dictionary with segment data indexed by segment name
    """

    t_pose = {'segments_counts': len(segment_names), 'segments': []}

    orientations = frame_element_as_floats(frame_element, 'orientation')
    offsets = frame_element_as_floats(frame_element, 'position')

    for index in range(len(segment_names)):
        segment = {'pos_g': get_3d_vector(offsets, index),
                   'q_gb': get_4d_vector(orientations, index)}
        t_pose['segments'].append(segment)

    return t_pose


def get_segment_data_from_frame(frame_element, segment_names):
    """
    Extract segment data from a frame

    :param frame_element: The frame element to process
    :param segment_names: a list with the segment names
    :return: a dictionary with segment data indexed by segment name
    """

    segment_data = collections.OrderedDict()

    orientations = frame_element_as_floats(frame_element, 'orientation')
    offsets = frame_element_as_floats(frame_element, 'position')
    velocities = frame_element_as_floats(frame_element, 'velocity')
    accelerations = frame_element_as_floats(frame_element, 'acceleration')
    angular_velocity = frame_element_as_floats(frame_element, 'angularVelocity')
    angular_acceleration = frame_element_as_floats(frame_element, 'angularAcceleration')

    for index in range(len(segment_names)):
        segment_name = segment_names[index]
        segment_data[segment_name] = collections.OrderedDict()
        segment_data[segment_name]['ori'] = get_4d_vector(orientations, index)
        segment_data[segment_name]['pos'] = get_3d_vector(offsets, index)
        segment_data[segment_name]['vel'] = get_3d_vector(velocities, index)
        segment_data[segment_name]['acc'] = get_3d_vector(accelerations, index)
        segment_data[segment_name]['ang_vel'] = get_3d_vector(angular_velocity, index)
        segment_data[segment_name]['ang_acc'] = get_3d_vector(angular_acceleration, index)

    center_of_mass = frame_element_as_floats(frame_element, 'centerOfMass')
    if center_of_mass:
        segment_data['com'] = {'pos': [], 'vel': [], 'acc': []}
        index = 0
        for com_field in segment_data['com']:
            segment_data['com'][com_field] = get_3d_vector(center_of_mass, index)
            index += 1

    return segment_data


def get_sensor_data_from_frame(frame_element, sensor_names):
    """
    Extract sensor data from a frame

    :param frame_element: The frame element to process
    :param sensor_names: a list with the segment names
    :return: a dictionary with sensor data indexed by sensor name
    """

    sensor_data = collections.OrderedDict()

    orientations = frame_element_as_floats(frame_element, 'sensorOrientation')
    free_accelerations = frame_element_as_floats(frame_element, 'sensorFreeAcceleration')
    magnetic_field = frame_element_as_floats(frame_element, 'sensorMagneticField')

    for index in range(len(sensor_names)):
        sensor_name = sensor_names[index]
        sensor_data[sensor_name] = collections.OrderedDict()
        sensor_data[sensor_name]["ori"] = get_4d_vector(orientations, index)
        sensor_data[sensor_name]["mag"] = get_4d_vector(magnetic_field, index)
        sensor_data[sensor_name]["acc"] = get_3d_vector(free_accelerations, index)

    return sensor_data


def get_finger_data_from_frame(frame_element, finger_segment_names):
    """
    Extract finger data from a frame

    :param frame_element: The frame element to process
    :param finger_segment_names: a list with the finger segment names
    :return: a dictionary with finger data indexed by finger name
    """

    finger_data = {'left': {}, 'right': {}}

    for side in finger_data:
        orientations = frame_element_as_floats(frame_element, 'orientationFingers' + side.capitalize())
        offsets = frame_element_as_floats(frame_element, 'positionFingers' + side.capitalize())

        for index in range(len(finger_segment_names[side])):
            finger_name = finger_segment_names[side][index]
            finger_data[side][finger_name] = collections.OrderedDict()
            finger_data[side][finger_name]["ori"] = get_4d_vector(orientations, index)
            finger_data[side][finger_name]["pos"] = get_3d_vector(offsets, index)

    return finger_data


def get_contact_data_from_frame(frame_element, foot_contact_def):
    """
    Extract contact data from a frame

    :param frame_element: The frame element to process
    :param foot_contact_def: a list with the foot contact definitions
    :return: a list with contacts
    """

    contact_data = []
    element_value = frame_element.find('mvn:footContacts', ns)
    if element_value is not None:
        contacts = [int(value) for value in element_value.text.split(' ')]

        for index in range(len(contacts)):
            if contacts[index] == 1:
                contact_data.append(foot_contact_def[index])

    return contact_data


def frame_element_as_floats(frame_element, element):
    """
    Find a named element in a frame element, extract the text from it, split that and return
    the values as an array of floats

    :param frame_element: The mvnx frame element to process
    :param element: The name of the sub element to find
    :return: an array of floating point values
    """

    element_value = frame_element.find('mvn:' + element, ns)
    return [float(value) for value in element_value.text.split(' ')] if element_value is not None else []


def get_4d_vector(raw_vector, index):
    return np.array(raw_vector[index * 4:index * 4 + 4])


def get_3d_vector(raw_vector, index):
    return np.array(raw_vector[index * 3:index * 3 + 3])


def init_file_data(mvnx_file):
    meta_data = {'version': '',
                 'original_filename': '',
                 'rec_date': '',
                 'name': '',
                 'color': '',
                 'comments': '',
                 'scenario': '',
                 'quality': '',
                 'sample_rate': 240}

    mvnx_file.file_data = {'segments': {},
                           'finger_segments': {'names': {'left': {}, 'right': {}},
                                               'elements': {'left': collections.OrderedDict(),
                                                            'right': collections.OrderedDict()}},
                           'sensors': {'names': [], 'elements': collections.OrderedDict()},
                           'joints': {'names': [], 'elements': collections.OrderedDict()},
                           'ergo_joints': {'names': [], 'elements': collections.OrderedDict()},
                           'finger_joints': {'names': {'left': {}, 'right': {}},
                                             'elements': {'left': collections.OrderedDict(),
                                                          'right': collections.OrderedDict()}},
                           'foot_contact_def': {},
                           'frames': {},
                           'tpose': {},
                           'tpose_isb': {},
                           'identity': {},
                           'meta_data': meta_data}



import numpy as np
from pathlib import Path
from biomechzoo.utils.version import get_biomechzoo_version


def set_zoosystem(fl=None):
    """
    Create the 'zoosystem' branch for data being imported to BiomechZoo.

    Args:
        fl: str
            Path to the source file (e.g., C3D or CSV).

    Returns:
        zoosystem: dict
            Dictionary containing default BiomechZoo system parameters.
    """

    # Default top-level fields
    zch = ['Analog', 'Anthro', 'AVR', 'CompInfo', 'SourceFile',
           'Units', 'Version', 'Video']

    # Initialize top-level dict
    zoosystem = {key: {} for key in zch}

    # Section-specific defaults
    section = ['Video', 'Analog']
    for sec in section:
        zoosystem[sec]['Channels'] = []
        zoosystem[sec]['Freq'] = []
        zoosystem[sec]['Indx'] = []
        zoosystem[sec]['ORIGINAL_START_FRAME'] = []
        zoosystem[sec]['ORIGINAL_END_FRAME'] = []
        zoosystem[sec]['CURRENT_START_FRAME'] = []
        zoosystem[sec]['CURRENT_END_FRAME'] = []

    # Processing and AVR defaults
    zoosystem['Processing'] = ''
    zoosystem['AVR'] = 0

    # Force plates defaults
    zoosystem['Analog']['FPlates'] = {
        'CORNERS': [],
        'NUMUSED': 0,
        'LOCALORIGIN': [],
        'LABELS': []
    }

    # Version and source file
    zoosystem['Version'] = get_biomechzoo_version()
    if fl is None:
        zoosystem['SourceFile'] = 'None'  # ensure string
    else:
        zoosystem['SourceFile'] = str(Path(fl))  # ensure string

    # Units defaults
    zoosystem['Units'] = {
        'Markers': 'mm',
        'Angles': 'deg',
        'Forces': 'N',
        'Moments': 'Nmm',
        'Power': 'W/kg',
        'Scalars': 'mm'
    }

    return zoosystem


import numpy as np
from biomechzoo.mvn.load_mvnx import load_mvnx
from biomechzoo.mvn.mvn import JOINTS, SEGMENTS
from biomechzoo.utils.set_zoosystem import set_zoosystem

def mvnx2zoo_data(fl):
    """ loads mvnx file from xsens"""

    mvnx_file = load_mvnx(fl)

    # create zoo data dict
    data = {'zoosystem': set_zoosystem()}
    # extract joint angle data (All JOINTS may not exist in a given dataset)
    for key, val in JOINTS.items():
        try:
            r = mvnx_file.get_joint_angle(joint=key)
            data[val] = {
                'line': np.array(r),
                'event': {}
            }
        except KeyError:
            print('joint {} does not exist, skipping'.format(val))

    # extract segment orientations (All SEGMENTS may not exist in a given dataset)
    for key, val in SEGMENTS.items():
        try:
            r = mvnx_file.get_sensor_ori(segment=key)

            data[val] = {
                'line': np.array(r),
                'event': {}
            }
        except KeyError:
            print('segment {} does not exist, skipping'.format(val))

    # get foot strike events
    data = _get_foot_strike_events(mvnx_file, data)

    # add meta information
    data = _get_meta_info(fl, mvnx_file, data)

    return data


def is_valid_for_zoo(val):
    """
    Returns True if the value is valid for a MATLAB-compatible zoo structure.
    """
    if val is None:
        return False
    if isinstance(val, list) and len(val) == 0:
        return False
    if isinstance(val, np.ndarray) and val.size == 0:
        return False
    return True


def _get_meta_info(fl, mvnx_file, data):
    # todo: add more, see mvnx_file object
    data['zoosystem']['Video']['Freq'] = int(mvnx_file.frame_rate)
    data['zoosystem']['mvnx_version'] = mvnx_file.version
    data['zoosystem']['mvnx_configuration'] = mvnx_file.configuration
    data['zoosystem']['recording_date'] = mvnx_file.recording_date
    data['zoosystem']['original_file_name'] = mvnx_file.original_file_name
    data['zoosystem']['frame_count'] = mvnx_file.frame_count
    return data


def _get_foot_strike_events(mvnx_file, data):
    RHeel = np.zeros(mvnx_file.frame_count)
    LHeel = np.zeros(mvnx_file.frame_count)

    for n in range(mvnx_file.frame_count):
        list_contact = mvnx_file.get_foot_contacts(n)
        for contact in list_contact:
            if contact['segment_index'] == 17:
                RHeel[n] = True
            elif contact['segment_index'] == 21:
                LHeel[n] = True

    hs_r = []
    hs_l = []
    for i in range(1, len(LHeel)):  # Start from 1 to avoid i-1 out-of-range
        if RHeel[i - 1] == 0 and RHeel[i] == 1:
            hs_r.append(i)
        if LHeel[i - 1] == 0 and LHeel[i] == 1:
            hs_l.append(i)

    # add to event branch of any channel
    if 'jL5S1' in data:
        ch = 'jL5S1'
    else:
        ch = next(iter(data))

    if hs_r:
        for i, rHS in enumerate(hs_r):
            data[ch]['event']['R_FS' + str(i + 1)] = [rHS, 0, 0]
    if hs_l:
        for i, lHS in enumerate(hs_l):
            data[ch]['event']['L_FS' + str(i + 1)] = [lHS, 0, 0]

    return data


if __name__ == '__main__':
    """ testing """
    import os
    from src.biomechzoo.utils.zplot import zplot
    # -------TESTING--------
    project_root = os.path.dirname(os.path.dirname(os.path.dirname(os.path.dirname(os.path.abspath(__file__)))))
    fl = os.path.join(project_root, 'data', 'other', 'Flat001.mvnx')
    data = mvnx2zoo_data(fl)
    zplot(data, 'jRightKnee')



from biomechzoo.biomech_ops.continuous_relative_phase_line import continuous_relative_phase_line
from biomechzoo.processing.addchannel_data import addchannel_data


def continuous_relative_phase_data(data, ch_dist, ch_prox):
    """ This function determines the CRP on a 0-180 scale, correcting for
           discontinuity in the signals >180.
    See Also phase_angle_data.py and phase_angle_line.py
    """

    data_new = data.copy()
    prox = data[ch_prox]['line']
    dist = data[ch_dist]['line']
    crp = continuous_relative_phase_line(dist, prox)
    data_new = addchannel_data(data_new, ch_new_name=ch_dist + '_' + ch_prox + '_' + 'crp', ch_new_data=crp)
    return data_new


if __name__ == '__main__':
    # -------TESTING--------
    import os
    from biomechzoo.utils.zload import zload
    from biomechzoo.utils.zplot import zplot
    # note: crp should be computed on phase angle data. Here we just demonstrate that it works.
    current_dir = os.path.dirname(os.path.abspath(__file__))
    project_root = os.path.dirname(current_dir)
    fl = os.path.join(project_root, 'data', 'other', 'HC032A18_exploded.zoo')
    data = zload(fl)
    data = continuous_relative_phase_data(data, ch_dist='RKneeAngles_x', ch_prox='RHipAngles_x')
    zplot(data, 'RKneeAngles_x_RHipAngles_x_crp')




def batchdisp(msg, level=1, verbose='none'):
    """ utility to control verbosity level during batch processing"""
    level = _normalize_verbose(level)
    verbose = _normalize_verbose(verbose)
    if verbose >= level:
        print(msg)

def _normalize_verbose(verbose):
    if isinstance(verbose, int):
        if verbose not in (0, 1, 2):
            raise ValueError("Integer verbose level must be 0 (none), 1 (minimal), or 2 (all)")
        return verbose
    elif isinstance(verbose, str):
        verbose_map = {'none': 0, 'minimal': 1, 'all': 2}
        if verbose.lower() not in verbose_map:
            raise ValueError("String verbose level must be 'none', 'minimal', or 'all'")
        return verbose_map[verbose.lower()]
    else:
        raise TypeError("Verbose must be an int (0–2) or str ('none', 'minimal', 'all')")





from biomechzoo.biomech_ops.filter_line import filter_line


def filter_data(data, ch, filt=None):
    """
    Filter one or more channels from a zoo data dictionary using specified filter parameters.

    Arguments
    ----------
    data : dict
        The zoo data dictionary containing signal channels.
    ch : str or list of str
        The name(s) of the channel(s) to filter.
    filt : dict, optional
        Dictionary specifying filter parameters. Keys may include:
        - 'ftype': 'butter' (default)
        - 'order': filter order (default: 4)
        - 'cutoff': cutoff frequency or tuple (Hz)
        - 'btype': 'low', 'high', 'bandpass', 'bandstop' (default: 'lowpass')

    Returns
    -------
    dict
        The updated data dictionary with filtered channels.
    """

    if filt is None:
        filt = {'ftype': 'butter',
                'order': 4,
                'cutoff': 10,
                'btype': 'lowpass',
                'filtfilt': True}

    if isinstance(ch, str):
        ch = [ch]

    # loop through all channels and filter
    for c in ch:
        if c not in data:
            raise KeyError('Channel {} not found in data'.format(c))

        if 'fs' not in filt:

            video_channels = data['zoosystem']['Video']['Channels']
            analog_channels = data['zoosystem']['Analog']['Channels']

            if c in analog_channels:
                filt['fs'] = data['zoosystem']['Analog']['Freq']
            elif c in video_channels:
                filt['fs'] =  data['zoosystem']['Video']['Freq']
            else:
                raise ValueError('Channel not analog or video')

        signal_raw = data[c]['line']
        signal_filtered = filter_line(signal_raw=signal_raw, filt=filt)
        data[c]['line'] = signal_filtered

    return data



def continuous_relative_phase_line(dist, prox):
    """ This function determines the CRP on a 0-180 scale, correcting for
       discontinuity in the signals >180.

    Arguments
    dist, ndarray: data of distal segment or joint
    prox, ndarray: data of proximal segment or joibt

    Returns
    crp, ndarray: continous relative phase betweeen dist and prox data
    """
    temp_CRP = abs(dist - prox)
    idx = temp_CRP > 180  # This corrects discontinuity in the data and puts everything on a 0-180 scale.
    temp_CRP[idx] = 360 - temp_CRP[idx]
    crp = temp_CRP
    return crp


if __name__ == '__main__':
    # -------TESTING--------
    import os
    from biomechzoo.utils.zload import zload
    from biomechzoo.biomech_ops.phase_angle_line import phase_angle_line
    from matplotlib import pyplot as plt
    # note: crp should be computed on phase angle data. Here we just demonstrate that it works.
    current_dir = os.path.dirname(os.path.abspath(__file__))
    project_root = os.path.dirname(current_dir)
    fl = os.path.join(project_root, 'data', 'other', 'HC032A18_exploded.zoo')
    data = zload(fl)
    knee = data['RKneeAngles_x']['line']
    hip = data['RHipAngles_x']['line']
    knee_pa = phase_angle_line(knee)
    hip_pa = phase_angle_line(hip)
    crp = continuous_relative_phase_line(knee_pa, hip_pa)
    plt.plot(crp)
    plt.show()



import os
import inspect
import time

from biomechzoo.imu.tilt_algoirthm import tilt_algorithm_data
from biomechzoo.utils.engine import engine  # assumes this returns .zoo files in folder
from biomechzoo.utils.zload import zload
from biomechzoo.utils.zsave import zsave
from biomechzoo.utils.batchdisp import batchdisp
from biomechzoo.utils.get_split_events import get_split_events
from biomechzoo.utils.split_trial import split_trial
from biomechzoo.conversion.c3d2zoo_data import c3d2zoo_data
from biomechzoo.conversion.table2zoo_data import table2zoo_data
from biomechzoo.conversion.mvnx2zoo_data import mvnx2zoo_data
from biomechzoo.processing.removechannel_data import removechannel_data
from biomechzoo.processing.renamechannel_data import renamechannel_data
from biomechzoo.processing.removeevent_data import removeevent_data
from biomechzoo.processing.explodechannel_data import explodechannel_data
from biomechzoo.processing.addevent_data import addevent_data
from biomechzoo.processing.partition_data import partition_data
from biomechzoo.processing.renameevent_data import renameevent_data
from biomechzoo.biomech_ops.normalize_data import normalize_data
from biomechzoo.biomech_ops.phase_angle_data import phase_angle_data
from biomechzoo.biomech_ops.continuous_relative_phase_data import continuous_relative_phase_data
from biomechzoo.biomech_ops.filter_data import filter_data

class BiomechZoo:
    def __init__(self, in_folder, inplace=False, subfolders=None, name_contains=None, verbose=0):
        self.verbose = verbose
        self.in_folder = in_folder
        self.verbose = verbose
        self.inplace = inplace               # choice to save processed files to new folder
        self.subfolders = subfolders         # only run processes on list in subfolder
        self.name_contains = name_contains   # only run processes on files with name_contains in file name

        batchdisp('BiomechZoo initialized', level=1, verbose=verbose)
        batchdisp('verbosity set to: {}'.format(verbose), level=1, verbose=verbose)
        batchdisp('root processing folder set to: {}'.format(self.in_folder), level=1, verbose=verbose)
        if name_contains is not None:
            batchdisp('only include files containing name_contains string: {}'.format(self.name_contains), level=1, verbose=verbose)
        if subfolders is not None:
            if type(subfolders) is list:
                batchdisp('only process files in subfolder(s):', level=1, verbose=verbose)
                for subfolder in self.subfolders:
                    batchdisp('{}'.format(os.path.join(self.in_folder, subfolder)), level=1, verbose=verbose)
            else:
                batchdisp('only process files in subfolder(s): {}'.format(os.path.join(self.in_folder, self.subfolders)), level=1, verbose=verbose)

        if inplace:
            batchdisp('Processing mode: overwrite (inplace=True) (each step will be applied to same folder)', level=1, verbose=verbose)
        else:
            batchdisp('Processing mode: backup (inplace=False)(each step will be applied to a new folder)', level=1, verbose=verbose)

    def _update_folder(self, out_folder, inplace, in_folder):
        """
        Utility to update self.in_folder if not inplace.

        Parameters:
        - out_folder (str or None): The output folder provided by user
        - inplace (bool): Whether processing is inplace
        - in_folder (str): The current input folder
        """
        if not inplace:
            # get full path for out_folder
            in_folder_path = os.path.dirname(in_folder)
            self.in_folder = os.path.join(in_folder_path, out_folder)

        batchdisp('all files saved to: {}'.format(self.in_folder ), level=1, verbose=self.verbose)

    def mvnx2zoo(self, out_folder=None, inplace=False):
        """ Converts all .mvnx files in the folder to .zoo format """
        start_time = time.time()
        verbose = self.verbose
        in_folder = self.in_folder
        if inplace is None:
            inplace = self.inplace
        fl = engine(in_folder, extension='.mvnx', name_contains=self.name_contains, subfolders=self.subfolders)
        for f in fl:
            batchdisp('converting mvnx to zoo for {}'.format(f), level=2, verbose=verbose)
            data = mvnx2zoo_data(f)
            f_zoo = f.replace('.mvnx', '.zoo')
            zsave(f_zoo, data, inplace=inplace, out_folder=out_folder, root_folder=in_folder)
        method_name = inspect.currentframe().f_code.co_name
        batchdisp('{} process complete for {} file(s) in {:.2f} secs'.format(method_name, len(fl), time.time() - start_time), level=1, verbose=verbose)
        # Update self.folder after  processing
        self._update_folder(out_folder, inplace, in_folder)

    def c3d2zoo(self, out_folder=None, inplace=None):
        """ Converts all .c3d files in the folder to .zoo format """
        start_time = time.time()
        from ezc3d import c3d
        verbose = self.verbose
        in_folder = self.in_folder
        if inplace is None:
            inplace = self.inplace
        fl = engine(in_folder, extension='.c3d', name_contains=self.name_contains, subfolders=self.subfolders)
        for f in fl:
            batchdisp('converting c3d to zoo for {}'.format(f), level=2, verbose=verbose)
            c3d_obj = c3d(f)
            data = c3d2zoo_data(c3d_obj)
            f_zoo = f.replace('.c3d', '.zoo')
            zsave(f_zoo, data, inplace=inplace, out_folder=out_folder, root_folder=in_folder)
        method_name = inspect.currentframe().f_code.co_name
        batchdisp('{} process complete for {} file(s) in {:.2f} secs'.format(method_name, len(fl), time.time() - start_time), level=1, verbose=verbose)
        # Update self.folder after  processing
        self._update_folder(out_folder, inplace, in_folder)

    def table2zoo(self, extension, out_folder=None, inplace=None, skip_rows=0, freq=None):
        """ Converts generic .csv file in the folder to .zoo format """
        start_time = time.time()
        verbose = self.verbose
        in_folder = self.in_folder

        if extension.startswith('.'):
            extension = extension[1:]

        if inplace is None:
            inplace = self.inplace
        fl = engine(in_folder, extension=extension, name_contains=self.name_contains, subfolders=self.subfolders)
        for f in fl:
            batchdisp('converting {} to zoo for {}'.format(extension, f), level=2, verbose=verbose)
            data = table2zoo_data(f, extension=extension, skip_rows=skip_rows, freq=freq)
            f_zoo = f.replace(extension, '.zoo')
            zsave(f_zoo, data, inplace=inplace, out_folder=out_folder, root_folder=in_folder)
        method_name = inspect.currentframe().f_code.co_name
        batchdisp('{} process complete for {} file(s) in {:.2f} secs'.format(method_name, len(fl), time.time() - start_time), level=1, verbose=verbose)
        # Update self.folder after  processing
        self._update_folder(out_folder, inplace, in_folder)

    def xls2zoo(self, out_folder=None, inplace=None):
        raise NotImplementedError('Use table2zoo instead')

    def csv2zoo(self, out_folder=None, inplace=None):
        raise NotImplementedError('Use table2zoo instead')

    def parquet2zoo(self, out_folder=None, inplace=None):
        raise NotImplementedError('Use table2zoo instead')

    def tilt_algorithm(self, chname_avert, chname_medlat, chname_antpost, out_folder=None, inplace=False):
        """ tilt correction for acceleration data """
        start_time = time.time()
        verbose = self.verbose
        in_folder = self.in_folder
        if inplace is None:
            inplace = self.inplace
        fl = engine(in_folder, name_contains=self.name_contains, subfolders=self.subfolders)
        for f in fl:
            batchdisp('tilt correction of acceleration channels for {}'.format(f), level=2, verbose=verbose)
            data = zload(f)
            data = tilt_algorithm_data(data, chname_avert, chname_medlat, chname_antpost)
            zsave(f, data, inplace=inplace, out_folder=out_folder, root_folder=in_folder)
        method_name = inspect.currentframe().f_code.co_name
        batchdisp(
            '{} process complete for {} file(s) in {:.2f} secs'.format(method_name, len(fl), time.time() - start_time),
            level=1, verbose=verbose)
        # Update self.folder after  processing
        self._update_folder(out_folder, inplace, in_folder)

    def phase_angle(self, ch, out_folder=None, inplace=None):
        """ computes phase angles"""
        start_time = time.time()
        verbose = self.verbose
        in_folder = self.in_folder
        if inplace is None:
            inplace = self.inplace
        fl = engine(in_folder, extension='.zoo', name_contains=self.name_contains, subfolders=self.subfolders)
        for f in fl:
            if verbose:
                batchdisp('computing phase angles for {}'.format(f), level=2, verbose=verbose)
            data = zload(f)
            data = phase_angle_data(data, ch)
            zsave(f, data, inplace=inplace, out_folder=out_folder, root_folder=in_folder)
        method_name = inspect.currentframe().f_code.co_name
        batchdisp('{} process complete for {} file(s) in {:.2f} secs'.format(method_name, len(fl), time.time() - start_time), level=1, verbose=verbose)

        # Update self.folder after  processing
        self._update_folder(out_folder, inplace, in_folder)

    def continuous_relative_phase(self, ch_prox, ch_dist, out_folder=None, inplace=None):
        """ computes CRP angles"""
        start_time = time.time()
        verbose = self.verbose
        in_folder = self.in_folder
        if inplace is None:
            inplace = self.inplace
        fl = engine(in_folder, extension='.zoo', name_contains=self.name_contains, subfolders=self.subfolders)
        for f in fl:
            if verbose:
                batchdisp('computing CRP angles between channel {} (prox) and {} (dist) for {}'.format(ch_prox, ch_dist, f), level=2, verbose=verbose)
            data = zload(f)
            data = continuous_relative_phase_data(data, ch_dist, ch_prox)
            zsave(f, data, inplace=inplace, out_folder=out_folder, root_folder=in_folder)
        method_name = inspect.currentframe().f_code.co_name
        batchdisp('{} process complete for {} file(s) in {:.2f} secs'.format(method_name, len(fl), time.time() - start_time), level=1, verbose=verbose)

        # Update self.folder after  processing
        self._update_folder(out_folder, inplace, in_folder)

    def split_trial_by_gait_cycle(self, first_event_name, out_folder=None, inplace=None):
        """ split by gait cycle according to event_name"""
        start_time = time.time()
        verbose = self.verbose
        in_folder = self.in_folder
        if inplace is None:
            inplace = self.inplace
        fl = engine(in_folder, extension='.zoo', name_contains=self.name_contains, subfolders=self.subfolders)
        for f in fl:
            f_name = os.path.splitext(os.path.basename(f))[0]
            data = zload(f)
            split_events = get_split_events(data, first_event_name)
            if split_events is None:
                print('no event {} found, saving original file'.format(first_event_name))
                zsave(f, data, inplace=inplace, out_folder=out_folder, root_folder=in_folder)
            else:
                for i, _ in enumerate(split_events[0:-1]):
                    fl_new = f.replace(f_name, f_name + '_' + str(i + 1))
                    start = split_events[i]
                    end = split_events[i + 1]
                    batchdisp('splitting by gait cycle from {} to {} for {}'.format(start, end, f), level=2,
                              verbose=verbose)
                    data_new = split_trial(data, start, end)
                    if data_new is not None:
                        zsave(fl_new, data_new, inplace=inplace, out_folder=out_folder, root_folder=in_folder)
        method_name = inspect.currentframe().f_code.co_name
        batchdisp('{} process complete for {} file(s) in {:.2f} secs'.format(method_name, len(fl), time.time() - start_time), level=1, verbose=verbose)

        # Update self.folder after  processing
        self._update_folder(out_folder, inplace, in_folder)


    # def mean_absolute_relative_phase_deviation_phase(self, channels, out_folder=None, inplace=None):
    #     verbose = self.verbose
    #     in_folder = self.in_folder
    #     if inplace is None:
    #         inplace = self.inplace
    #
    #     fl = engine(in_folder)
    #     for f in fl:
    #         for channel in channels:
    #             batchdisp('collecting trials for marp and dp for {}'.format(f), level=2, verbose=verbose)
    #             data = zload(f)
    #             data = removechannel_data(data, ch, mode)
    #             zsave(f, data, inplace=inplace, root_folder=in_folder, out_folder=out_folder)
    #             batchdisp('remove channel complete', level=1, verbose=verbose)
    #
    #     # Update self.folder after  processing
    #     self._update_folder(out_folder, inplace, in_folder)
    def renameevent(self, evt, nevt, out_folder=None, inplace=None):
        """ renames event evt to nevt in all zoo files """
        start_time = time.time()
        verbose = self.verbose
        in_folder = self.in_folder
        if inplace is None:
            inplace = self.inplace
        fl = engine(in_folder, extension='.zoo', name_contains=self.name_contains, subfolders=self.subfolders)
        for f in fl:
            batchdisp('renaming events from {} to {} for {}'.format(evt, nevt ,f), level=2, verbose=verbose)
            data = zload(f)
            data = renameevent_data(data, evt, nevt)
            zsave(f, data, inplace=inplace, out_folder=out_folder, root_folder=in_folder)
        method_name = inspect.currentframe().f_code.co_name
        batchdisp('{} process complete for {} file(s) in {:.2f} secs'.format(method_name, len(fl), time.time() - start_time), level=1, verbose=verbose)

        # Update self.folder after  processing
        self._update_folder(out_folder, inplace, in_folder)

    def renamechannnel(self, ch, ch_new, out_folder=None, inplace=None):
        """ renames channels from ch to ch_new in all zoo files """
        start_time = time.time()
        verbose = self.verbose
        in_folder = self.in_folder
        if inplace is None:
            inplace = self.inplace
        fl = engine(in_folder, extension='.zoo', name_contains=self.name_contains, subfolders=self.subfolders)
        for f in fl:
            batchdisp('renaming channels from {} to {} for {}'.format(ch, ch_new ,f), level=2, verbose=verbose)
            data = zload(f)
            data = renamechannel_data(data, ch, ch_new)
            zsave(f, data, inplace=inplace, out_folder=out_folder, root_folder=in_folder)
        method_name = inspect.currentframe().f_code.co_name
        batchdisp('{} process complete for {} file(s) in {:.2f} secs'.format(method_name, len(fl), time.time() - start_time), level=1, verbose=verbose)

        # Update self.folder after  processing
        self._update_folder(out_folder, inplace, in_folder)

    def removechannel(self, ch, mode='remove', out_folder=None, inplace=None):
        """ removes channels from zoo files """
        start_time = time.time()
        verbose = self.verbose
        in_folder = self.in_folder
        if inplace is None:
            inplace = self.inplace
        fl = engine(in_folder, extension='.zoo', name_contains=self.name_contains, subfolders=self.subfolders)
        for f in fl:
            batchdisp('removing channels for {}'.format(f), level=2, verbose=verbose)
            data = zload(f)
            data = removechannel_data(data, ch, mode)
            zsave(f, data, inplace=inplace, out_folder=out_folder, root_folder=in_folder)
        method_name = inspect.currentframe().f_code.co_name
        batchdisp('{} process complete for {} file(s) in {:.2f} secs'.format(method_name, len(fl), time.time() - start_time), level=1, verbose=verbose)

        # Update self.folder after  processing
        self._update_folder(out_folder, inplace, in_folder)


    def removeevent(self, events, mode='remove', out_folder=None, inplace=None):
        """ removes channels from zoo files """
        start_time = time.time()
        verbose = self.verbose
        in_folder = self.in_folder
        if inplace is None:
            inplace = self.inplace
        fl = engine(in_folder, extension='.zoo', name_contains=self.name_contains, subfolders=self.subfolders)
        for f in fl:
            batchdisp('removing events {} for {}'.format(events, f), level=2, verbose=verbose)
            data = zload(f)
            data = removeevent_data(data, events, mode)
            zsave(f, data, inplace=inplace, out_folder=out_folder, root_folder=in_folder)
        method_name = inspect.currentframe().f_code.co_name
        batchdisp('{} process complete for {} file(s) in {:.2f} secs'.format(method_name, len(fl), time.time() - start_time), level=1, verbose=verbose)

        # Update self.folder after  processing
        self._update_folder(out_folder, inplace, in_folder)


    def explodechannel(self, out_folder=None, inplace=None):
        """ explodes all channels in a zoo file """
        start_time = time.time()
        verbose = self.verbose
        in_folder = self.in_folder
        if inplace is None:
            inplace = self.inplace
        fl = engine(in_folder, extension='.zoo', name_contains=self.name_contains, subfolders=self.subfolders)
        for f in fl:
            if verbose:
                batchdisp('removing channels for {}'.format(f), level=2, verbose=verbose)
            data = zload(f)
            data = explodechannel_data(data)
            zsave(f, data, inplace=inplace, out_folder=out_folder, root_folder=in_folder)
        method_name = inspect.currentframe().f_code.co_name
        batchdisp('{} process complete for {} file(s) in {:.2f} secs'.format(method_name, len(fl), time.time() - start_time), level=1, verbose=verbose)

        # Update self.folder after  processing
        self._update_folder(out_folder, inplace, in_folder)

    def normalize(self, nlen=101, out_folder=None, inplace=None):
        """ time normalizes all channels to length nlen """
        start_time = time.time()
        verbose = self.verbose
        in_folder = self.in_folder
        if inplace is None:
            inplace = self.inplace
        fl = engine(in_folder, extension='.zoo', name_contains=self.name_contains, subfolders=self.subfolders)
        for f in fl:
            if verbose:
                batchdisp('normalizing channels to length {} for {}'.format(nlen, f), level=2, verbose=verbose)
            data = zload(f)
            data = normalize_data(data, nlen)
            zsave(f, data, inplace=inplace, out_folder=out_folder, root_folder=in_folder)
        method_name = inspect.currentframe().f_code.co_name
        batchdisp('{} process complete for {} file(s) in {:.2f} secs'.format(method_name, len(fl), time.time() - start_time), level=1, verbose=verbose)

        # Update self.folder after  processing
        self._update_folder(out_folder, inplace, in_folder)

    def addevent(self, ch, event_type, event_name, out_folder=None, inplace=None, constant=None):
        """ adds events of type evt_type with name evt_name to channel ch """
        start_time = time.time()
        verbose = self.verbose
        in_folder = self.in_folder
        if inplace is None:
            inplace = self.inplace
        fl = engine(in_folder, extension='.zoo', name_contains=self.name_contains, subfolders=self.subfolders)
        for f in fl:
            if verbose:
                batchdisp('adding event {} to channel {} for {}'.format(event_type, ch, f), level=2, verbose=verbose)
            data = zload(f)
            data = addevent_data(data, ch, event_type, event_name, constant)
            zsave(f, data, inplace=inplace, out_folder=out_folder, root_folder=in_folder)
        method_name = inspect.currentframe().f_code.co_name
        batchdisp('{} process complete for {} file(s) in {:.2f} secs'.format(method_name, len(fl), time.time() - start_time), level=1, verbose=verbose)

        # Update self.folder after  processing
        self._update_folder(out_folder, inplace, in_folder)

    def partition(self, evt_start, evt_end, out_folder=None, inplace=None):
        """ partitions data between events evt_start and evt_end """
        start_time = time.time()
        verbose = self.verbose
        in_folder = self.in_folder
        if inplace is None:
            inplace = self.inplace
        fl = engine(in_folder, extension='.zoo', name_contains=self.name_contains, subfolders=self.subfolders)
        for f in fl:
            if verbose:
                batchdisp('partitioning data between events {} and {} for {}'.format(evt_start, evt_end, f), level=2, verbose=verbose)
            data = zload(f)
            data = partition_data(data, evt_start, evt_end)
            zsave(f, data, inplace=inplace, out_folder=out_folder, root_folder=in_folder)
        method_name = inspect.currentframe().f_code.co_name
        batchdisp('{} process complete for {} file(s) in {:.2f} secs'.format(method_name, len(fl), time.time() - start_time), level=1, verbose=verbose)
        # Update self.folder after  processing
        self._update_folder(out_folder, inplace, in_folder)

    def filter(self, ch, filt=None, out_folder=None, inplace=None):
        """ filter data"""
        start_time = time.time()
        verbose = self.verbose
        in_folder = self.in_folder
        if inplace is None:
            inplace = self.inplace
        fl = engine(in_folder, name_contains=self.name_contains, subfolders=self.subfolders)
        for f in fl:
            if verbose:
                batchdisp('filtering data for channel {} in {}'.format(ch, f), level=2, verbose=verbose)
            data = zload(f)
            data = filter_data(data, ch, filt)
            zsave(f, data, inplace=inplace, out_folder=out_folder, root_folder=in_folder)
        method_name = inspect.currentframe().f_code.co_name
        batchdisp('{} process complete for {} file(s) in {:.2f} secs'.format(method_name, len(fl), time.time() - start_time),
            level=1, verbose=verbose)
        # Update self.folder after  processing
        self._update_folder(out_folder, inplace, in_folder)




from biomechzoo.utils.findfield import findfield
import warnings
import copy
import numpy as np


def partition_data(data, evt_start, evt_end):
    """ partition data for all channels between events evt_start and evt_end"""

    # extract event values
    e1, _ = findfield(data, evt_start)
    e2, _ = findfield(data, evt_end)

    if e1 is None or e2 is None or len(e1) == 0 or len(e2) == 0:
        raise ValueError(f"Event not found: evt_start='{evt_start}' returned {e1}, evt_end='{evt_end}' returned {e2}")

    # convert to int and get first value
    e1 = int(e1[0])
    e2 = int(e2[0])

    data_new = copy.deepcopy(data)
    for ch_name, ch_data in sorted(data_new.items()):
        if ch_name != 'zoosystem':
            r = ch_data['line']
            try:
                if r.ndim == 1:
                    data_new[ch_name]['line'] = r[e1:e2]
                else:
                    data_new[ch_name]['line'] = r[e1:e2, :]
            except (IndexError, ValueError) as e:
                # IndexError: if e1[0]:e2[0] goes beyond the available indices
                # ValueError: less likely, but may arise with shape mismatches
                warnings.warn(f"Skipping {ch_name} due to error: {e}")

            # partition events
            events = ch_data['event']
            for event_name, value in events.items():
                original_frame = int(value[0])
                if original_frame == 999:
                    continue  # do not change outlier markers
                else:
                    arr = np.array(data_new[ch_name]['event'][event_name], dtype=np.int32)
                    arr[0] = original_frame - e1
                    data_new[ch_name]['event'][event_name] = arr

    return data_new



import numpy as np


def compute_sampling_rate_from_time(t, verbose=False):
    """ computes sampling rate from time column
    Arguments
    t, 1D numpy array, recorded times

    Returns
    fsamp, int: sampling rate of capture
    """
    # Calculate differences between consecutive time points
    dt = np.diff(t)

    # Average time difference (seconds)
    avg_dt = np.mean(dt)

    # Sampling frequency (Hz)
    # Sampling frequency (Hz)
    fsamp = int(np.round(1 / avg_dt))

    if verbose:
        print('Inferred sampling rate: {} Hz'.format(fsamp))

    return fsamp



import pandas as pd
import os
import re

from biomechzoo.utils.set_zoosystem import set_zoosystem
from biomechzoo.utils.compute_sampling_rate_from_time import compute_sampling_rate_from_time


def table2zoo_data(fl, extension, skip_rows=0, freq=None):

    if extension == 'csv':
        df, metadata, freq = _csv2zoo(fl, skip_rows=skip_rows, freq=freq)

    elif extension == 'parquet':
        df, metadata, freq = _parquet2zoo(fl, skip_rows=skip_rows, freq=freq)
    else:
        raise ValueError('extension {} not implemented'.format(extension))

    # assemble zoo data
    data = {'zoosystem': set_zoosystem()}
    for ch in df.columns:
        data[ch] = {
            'line': df[ch].values,
            'event': []
        }


    # now try to calculate freq from a time column
    if freq is None:
        time_col = [col for col in df.columns if 'time' in col.lower()]
        if time_col is not None and len(time_col) > 0:
            time_data = df[time_col].to_numpy()[:, 0]
            freq = compute_sampling_rate_from_time(time_data)
        else:
            raise ValueError('Unable to compute sampling rate for time column, please specify a sampling frequency'
                             )
    # add metadata
    data['zoosystem']['Video']['Freq'] = freq
    data['zoosystem']['Analog']['Freq'] = 'None'

    if metadata is not None:
        data['zoosystem']['Other'] = metadata

    return data


def _parquet2zoo(fl, skip_rows=0, freq=None):
    df = pd.read_parquet(fl)
    metadata = None
    return df, metadata, freq

def _csv2zoo(fl, skip_rows=0, freq=None):
    header_lines = []
    with open(fl, 'r') as f:
        for line in f:
            header_lines.append(line.strip())
            if line.strip().lower() == 'endheader':
                break
    # Parse metadata
    metadata = _parse_metadata(header_lines)

    # try to find frequency in metadata
    if freq is None:
        if 'freq' in metadata:
            freq = metadata['freq']
        elif 'sampling_rate' in metadata:
            freq = metadata['sampling_rate']
        else:
            freq = None  # or raise an error

    # read csv
    df = pd.read_csv(fl, skiprows=skip_rows)

    return df, metadata, freq




def _parse_metadata(header_lines):
    metadata = {}
    for line in header_lines:
        if '=' in line:
            key, val = line.split('=', 1)
            key = key.strip()
            val = val.strip()

            # Strip trailing commas and whitespace explicitly
            val = val.rstrip(',').strip()

            # Extract first numeric token if any
            match = re.search(r'[-+]?\d*\.?\d+', val)
            if match:
                num_str = match.group(0)
                try:
                    val_num = int(num_str)
                except ValueError:
                    val_num = float(num_str)
            else:
                # Now val should be clean of trailing commas, so just lower case it
                val_num = val.lower()

            metadata[key] = val_num
    return metadata




if __name__ == '__main__':
    """ for unit testing"""
    current_dir = os.path.dirname(os.path.abspath(__file__))
    project_root = os.path.dirname(current_dir)
    csv_file = os.path.join(project_root, 'data', 'other', 'opencap_walking1.csv')

    data = table2zoo_data(csv_file)



def renameevent_data(data, evt, nevt):
    """
    Rename events in the Zoo data structure.

    Parameters
    ----------
    data : dict
        The Zoo-formatted dictionary.
    evt : str or list of str
        Names of existing events to rename.
    nevt : str or list of str
        Names of new events to apply.

    Returns
    -------
    data : dict
        Updated Zoo data with renamed events.
    """
    # Convert to list if passed as single string
    if isinstance(evt, str):
        evt = [evt]
    if isinstance(nevt, str):
        nevt = [nevt]

    if len(evt) != len(nevt):
        raise ValueError("`evt` and `nevt` must have the same length.")

    # Get all data channels except 'zoosystem'
    channels = [ch for ch in data if ch != 'zoosystem']
    for old_name, new_name in zip(evt, nevt):
        for ch in channels:
            events = data[ch].get('event', {})
            if old_name in events:
                data[ch]['event'][new_name] = events[old_name]
                del data[ch]['event'][old_name]

    return data


if __name__ == '__main__':
    # -------TESTING--------
    import os
    from src.biomechzoo.utils.zload import zload
    # get path to sample zoo file
    current_dir = os.path.dirname(os.path.abspath(__file__))
    project_root = os.path.dirname(current_dir)
    fl = os.path.join(project_root, 'data', 'other', 'HC030A05.zoo')

    # load  zoo file
    data = zload(fl)
    evt = ['Left_FootStrike1', 'Left_FootStrike2', 'NonExistingEvent']

    # see existing keys
    missing = [k for k in evt if k not in data['SACR']['event']]
    print("Missing keys:", missing)   # expected behavior is missing 'NonExistingEvent'

    nevt = ['LFS_1', 'LFS2', 'NE_1']
    data = renameevent_data(data, evt=evt, nevt=nevt)

    # after applying renameevent_data
    missing = [k for k in nevt if k not in data['SACR']['event']]
    print("Missing keys:", missing)   # expected behavior is missing 'NE_1'



import os
import numpy as np


def engine(root_folder, extension='.zoo', subfolders=None, name_contains=None, verbose=False):
    """
    Recursively search for files with a given extension, optionally filtering by
    specific subfolders and substrings in filenames.

    This function walks through every directory and subdirectory starting at
    'root_folder'. If 'subfolders' is provided, it limits the search only to those
    folders (or any of their subfolders) whose names appear in 'subfolders'. For
    example, if subfolders=['Straight'], it will only consider files inside any
    folder named 'Straight' at any depth within the root folder.

    For each file found, it checks whether the file's extension matches the given
    'extension' (case-insensitive). If 'name_contains' is specified, it also
    requires the filename to contain that substring (case-insensitive).

    Arguments:
        root_folder (str): The root directory path where the search begins.
        extension (str): File extension to search for (e.g., '.zoo', '.c3d'). Default .zoo
        subfolders (list or str, optional): List of folder names to restrict the search to.
            Only files inside these folders (or their subfolders) are included.
            If None, search all subfolders.
        name_contains (str, or list; optional): Substring that must be present in the filename
            (case-insensitive). If None, no substring filtering is applied.
        verbose (bool, optional): If true, displays additional information to user
    Returns:
        list of str: List of full file paths matching the criteria.
    """
    # check format of subfolder (string or list)
    if subfolders is not None:
        if type(subfolders) is str:
            subfolders = [subfolders]

    # check format of name_contants (str or list)
    if name_contains is not None:
        if type(name_contains) is str:
            name_contains = [name_contains]

    matched_files = []

    subfolders_set = set(subfolders) if subfolders else None
    for dirpath, _, filenames in os.walk(root_folder):
        if subfolders_set is not None:
            rel_path = os.path.relpath(dirpath, root_folder)
            if rel_path == '.':
                continue
            # Split the relative path into all folder parts
            parts = rel_path.split(os.sep)
            # Check if any folder in the path matches one in subfolders_set
            if not any(part in subfolders_set for part in parts):
                continue

        for file in filenames:
            if not file.lower().endswith(extension.lower()):
                continue
            full_path = os.path.join(dirpath, file)
            if name_contains is not None:
                match = False
                for name_contain in name_contains:
                    if name_contain and name_contain.lower() in full_path.lower():  # <-- check full path
                        match = True
                        break
                if not match:
                    continue
            matched_files.append(full_path)

    # sort list
    matched_files = np.sort(matched_files)

    if verbose:
        print("Found {} {} files in subfolder(s) named {} with substring {}:".format(len(matched_files), extension, subfolders, name_contains))
        for f in matched_files:
            print('{}'.format(f))

    return matched_files


if __name__ == '__main__':
    """ testing: use engine to search for files in any subfolder called 'Straight' for files with the substring 'HC03'
    with extension .c3d in the sample study folder (data)"""
    # -------TESTING--------
    current_dir = os.path.dirname(os.path.abspath(__file__))
    project_root = os.path.dirname(os.path.dirname(os.path.dirname(current_dir)))
    sample_dir = os.path.join(project_root, 'data', 'sample_study', 'raw c3d files')
    c3d_files = engine(sample_dir, '.c3d', subfolders=['Straight'], name_contains='HC03', verbose=True)
